/*
 * Gabadec - A dot-bag file decoder.
 * (C) 2003 Jean Delvare <khali@linux-fr>
 *
 * Licensed under the GNU Public license.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *   For the avoidance of doubt the "preferred form" of this code is one which
 *   is in an open unpatent encumbered format. Where cryptographic key signing
 *   forms part of the process of creating an executable the information 
 *   including keys needed to generate an equivalently functional executable
 *   are deemed to be part of the source code.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define  VERSION   "1.0"
#define  GABA      0x41424147
#define  BUFSIZE   64

typedef struct
{
	unsigned long signature;
	unsigned long revision;
	unsigned long count;
	unsigned long length;
} gabaheader;

typedef struct
{
	char name[32];
	unsigned long offset;
	unsigned long length;
	unsigned long srate;
	unsigned long type;
	unsigned long subtype;
	unsigned long padding[3];
} gabaentry;

int extrackt(FILE* inf, FILE* outf, unsigned int l, char* buffer )
{
	unsigned int w=0;

	while(w<l)
	{
		unsigned int s, r;
		s=(l-w>BUFSIZE?BUFSIZE:l-w);
		r=fread(buffer, 1, s, inf);
		if(r<s)
		{
			perror("fread");
			return -1;
		}

		if(fwrite(buffer, 1, r, outf)<r)
		{
			perror("fwrite");
			return -1;
		}
		w+=r;
	}

	return 0;
}

int main(int argc, char** argv)
{
	FILE* f;
	const char* filename="MUSIC.GAB";
	gabaheader gh;
	gabaentry ge;
	unsigned int i;
	char* buffer;
	
	printf("Gabadec %s\n", VERSION);
	printf("(C) 2003 Jean Delvare <khali@linux-fr.org>\n"); 
	
	if(argc>=2)
		filename=argv[1];
	printf("Extracting tracks from file \"%s\".\n", filename);

	printf("Opening source file...");
	f=fopen(filename, "rb");
	if(f==NULL)
	{
		printf(" failed.\n");
		perror("fopen");
		return 1;
	}
	printf(" OK.\n");

	printf("Reading header...");
	if(fread(&gh, sizeof(gh), 1, f)<1)
	{
		printf(" failed.\n");
		perror("fread");
		fclose(f);
		return 1;
	}
	printf(" OK.\n");

	printf("Checking file signature...");
	if(gh.signature!=GABA)
	{
		printf(" failed.\n");
		fclose(f);
		return 1;
	}
	printf(" OK.\n");

	printf("File properties:\n");
	printf("  revision %lu\n", gh.revision);
	printf("  %lu entries of %lu bytes each\n", gh.count, gh.length);

	if(gh.revision!=4)
		printf("Warning, untested revision.\n");
	if(gh.length<sizeof(ge))
	{
		printf("Records are too short, exiting.\n");
		fclose(f);
		return 2;
	}

	buffer=(char*)malloc(BUFSIZE*1024*sizeof(char));
	if(buffer==NULL)
	{
		perror("malloc");
		fclose(f);
		return 1;
	}

	for(i=0;i<gh.count;i++)
	{
		char outname[sizeof(ge.name)+5];
		FILE* of;

		if(fseek(f, sizeof(gh)+i*gh.length, SEEK_SET)==-1)
		{
			perror("fseek");
			fclose(f);
			return 1;
		}
		if(fread(&ge, sizeof(ge), 1, f)<1)
		{
			perror("fread");
			fclose(f);
			return 1;
		}

		if(ge.type!=0x24 && ge.type!=0x25)
		{
			printf("Skipping non-mp3 entry (type=0x%02lx, subtype=0x%08lx).\n", ge.type, ge.subtype);
		}
		else
		{
			printf("Seeking to track %u (type=0x%02lx)...", i+1, ge.type);
			if(fseek(f, ge.offset, SEEK_SET)==-1)
			{
				printf(" failed.\n");
				perror("fseek");
				fclose(f);
				return 1;
			}
			printf(" OK.\n");

			strncpy(outname, ge.name, sizeof(ge.name));
			outname[sizeof(ge.name)]='\0';
			strcat(outname, ".mp3");

			printf("Creating file \"%s\"...", outname);
			of=fopen(outname, "wb");
			if(of==NULL)
			{
				printf(" failed.\n");
				perror("fopen");
				fclose(f);
				return 1;
			}
			printf(" OK.\n");

			if(extrackt(f, of, ge.length, buffer)!=0)
			{
				fclose(of);
				fclose(f);
				return 1;
			}

			if(fclose(of)!=0)
			{
				perror("fclose");
				fclose(f);
				return 1;
			}
		}
	}

	free(buffer);

	printf("Closing source file...");
	if(fclose(f)!=0)
	{
		printf(" failed.\n");
		perror("fclose");
		return 1;
	}
	printf(" OK.\n");

	return 0;
}
